#!/usr/bin/python

import sys
import asmclasses
from asmconsts import *

import x86ops as CurrentCpu

try:
    import cPickle as pickle
except:
    try:
        import pickle as pickle
    except:
        print "Error:", sys.exc_info()[1]

CURRENT_CPU = "x86"

SYMBOL_LIST = {}

def parseAsmLine(line):
    """ Return the command and a list of operands in the order they appear """
    
    global SYMBOL_LIST
    
    instruction = ""
    operands = []

    foundParenthesys = False
    foundComa = False
    buf = ""
    instruction = ""

    for c in line.code:
        #print "Char is", c
        if c in ["\t", " "]:
            if buf.strip(" ").strip("\t") != "":
                if instruction == "":
                    instruction = buf
                else:
                    buf = buf.strip("<").strip(">")
                    operands.append(buf)

                buf = ""
            continue

        if c == "(":
            foundParenthesys = True

        if c == ")":
            foundParenthesys = False
        
        if c == ",":
            foundComa = True

            if foundParenthesys == False:
                buf = buf.strip("<").strip(">")
                operands.append(buf)
                buf = ""
                continue

        buf += c

    if instruction == "":
        instruction = buf
    else:
        buf = buf.strip("<").strip(">")
        operands.append(buf)

    i = -1
    for op in operands:
        i += 1
        if op.find("$0x") > -1 and line.description is not None:
            SYMBOL_LIST[op] = repr(line.description)
        operands[i] = getSymbolName(op)
    
    return instruction, operands

def getDatabase(dbname):
    f = file(dbname, "r")
    
    #
    # The database you will load will be in the following format
    #
    # 1) Raw data found in the .rodata section
    # 2) The whole program in Python structures
    #
    return pickle.load(f)

def getRegisterValue(op, registersList):

    if registersList.has_key(op):
        return registersList[op]
    else:
        return op

def printFunctionStart(procname, flow):

    funcStartcode = "int " + procname + "("
    numArgs = 0

    for arg in flow.localArgs:
        numArgs += 1
        if numArgs == len(flow.localArgs):
            funcStartcode += "int arg" + str(numArgs)
        else:
            funcStartcode += "int arg" + str(numArgs) + ", "

    funcStartcode += ")"
    return funcStartcode

def getSymbolList(flow):
    global SYMBOL_LIST

    i = 0
    for args in flow.localArgs:
        i += 1
        for arg in args:
            SYMBOL_LIST[arg + "(%ebp)"] = "arg" + str((eval(arg) / 4)-1)

    i = 0
    for locals in flow.vars:
        i += 1
        for arg in locals:
            SYMBOL_LIST[arg + "(%ebp)"] = "local" + str(((0xffffffff - eval(arg))/4)+1)

def getSymbolName(symbol):

    global SYMBOL_LIST

    if SYMBOL_LIST.has_key(symbol):
        local = SYMBOL_LIST[symbol]
    else:
        local = symbol
        
    if local.endswith("@plt"):
        local = local.rstrip("@plt")
    
    return local
    
def getOperands(opList):
    j = -1
    ops = opList
    for op in opList:
        j += 1
        ops[j] = getSymbolName(op)
    
    return ops

def processAritmethicOperation(ins, ops, registersList):
    
    maxOps = len(ops)-1
    assignToOp = ops[maxOps]

    if ins == "add":
        if registersList.has_key(assignToOp):
            registersList[assignToOp] = registersList[assignToOp] + "+" + toHumanValue(ops[0])
        else:
            operand = getSymbolName(ops[0])
            return "%s = %s + %s;" % (assignToOp, operand, assignToOp)

def checkContinue(ins, ops):
    if ins == "sub":
        if len(ops) > 1:
            if ops[1] == "%esp":
                return True
            else:
                return False
        else:
            return False
    else:
        return False

    return True

def toHumanValue(val):

    val = val.lstrip("$")
    
    if val.startswith("0x") and val.isalnum():
        val = str(eval(val))

    return val

def isCommonRegister(reg):
    return reg.lower().strip("%") in ["eax", "ebx", "ecx", "edx", "edi", "esi", "ebp", "esp"]

def decompile(dbname, procname):

    rodata, obj = getDatabase(dbname)
    flow = getFlow(rodata, obj, procname)
    getSymbolList(flow)
    #print "***Second pass!"
    argsQueue = []
    registersList = {}
    printFinalBracket = False

    for proc in obj:
        if proc.name == procname:
            printFinalBracket = True
            print printFunctionStart(procname, flow)
            print "{"
            
            level = 1
            i = 0
            foundStartCode = False
            previousInstruction = ""
            previousLine = ""
            
            if flow.localVariables:
                for i in range(0, flow.localVariables):
                    print "int local%d;" % (i+1)
                print

            for line in proc.lines:
                i += 1
                ins, ops = parseAsmLine(line)

                if i == 1 and ins == "push":
                    continue
                elif i == 2 and ins == "mov":
                    continue
                elif i == 3 and ins == "sub":
                    continue

                if procname == "main":
                    if not foundStartCode:
                        if not checkContinue(ins, ops):
                            continue
                        else:
                            foundStartCode = True
                            continue

                #print "//", line.code, ";", line.description
                ops = getOperands(ops)
                
                if line.code.replace(" ", "") in IGNORE_INSTRUCTIONS_DES[CURRENT_CPU]:
                    #print "Silently ignoring %s" % line.code
                    continue

                previousInstruction = ins
                
                if ins.startswith("push"):
                    if registersList.has_key(ops[0]):
                        val = registersList[ops[0]]
                    else:
                        val = ops[0]

                    val = toHumanValue(val)
                    argsQueue.append(val)
                elif ins.startswith("mov"):
                    #print registersList
                    if ops[len(ops) -1].find("(%esp)") > -1:
                        val = getRegisterValue(ops[0], registersList)
                        argsQueue.append(val)
                        continue
                    else:
                        registersList[ops[1]] = getRegisterValue(ops[0], registersList)
                    
                    strLine = "%s = %s;" % (ops[1], getRegisterValue(ops[0], registersList))
                    
                    if isCommonRegister(ops[0]):
                        print "\t"*level, strLine
                elif ins in ["add", "sub", "shr", "shl"]:
                    data = processAritmethicOperation(ins, ops, registersList)
                    
                    if data:
                        print "\t"*level, data
                elif ins == "call":
                    j = 0
                    strLine = ""

                    for arg in argsQueue[::-1]:
                        j += 1
                        
                        if type(arg) is list:
                            arg = getSymbolName(arg[0])
                        else:
                            arg = getSymbolName(arg)
                        
                        if j == 1:
                            strLine = arg
                        else:
                            strLine += ", " + arg

                    strLine = "(%s);" % strLine
                    argsQueue = []
                    previousLine = ops[1] + strLine
                    
                    print "\t"*level, ops[1] + strLine
                elif ins == "ret":
                    strLine = ""
                    if i > 0 and previousInstruction.find("call") > -1:
                        strLine = previousLine
                    elif registersList.has_key("%eax"):
                        strLine = toHumanValue(registersList["%eax"])

                    if len(ops) == 0:
                        print "\t"*level, "return(%s);" % strLine
                    continue
                else:
                    print "\t"*level, line.code

    if printFinalBracket:
        print "}"
        print

def getFlow(rodata, obj, procname):

    #print "***First pass!"
    cpu = CurrentCpu.CVariableFinder()
    flow = asmclasses.CProcTree()
    CurrentCpu.GLOBAL_RODATA = rodata


    for proc in obj:
        if proc.name == procname:
            i = 0
            
            functionProlog = False
            tested = 0
            block = []

            for line in proc.lines:
                i += 1
                block.append(line.code)
                # Current Instruction (Assume is x86)
                instructions = line.code.split(" ")

                if functionProlog:
                    if tested == 1:
                        tested = 0
                        functionProlog = False
                    else:
                        tested = 1
    
                if i == len(CurrentCpu.FUNCTION_PROLOG):
                    if cpu.checkCondition(block, CurrentCpu.FUNCTION_PROLOG):
                        functionProlog = True
                        #print "### FUNCTION PROLOG ###"

                if line.code.startswith("j") or line.code.startswith("lj") or line.code.find("call") == 0:
                    pass
                else:
                    cpu.analyzeInstruction(line.code, flow)
                    
                if functionProlog:
                    if instructions[0] == CurrentCpu.GET_SPACE_FOR_VARS:
                        space = cpu.getVariableSpace(line.code)
                        #print "### Space reserved %s ###" % space
                        #print "### So number of locals appears to be %d ###" % (int(space) / 8)
                        flow.localVariables = int(space) / 8

    """
    print "Expected number of local variables + local parameters", flow.localVariables
    print "Foreign arguments",flow.args
    print "Local arguments", flow.localArgs
    print "Local vars", flow.vars
    """

    return flow

def usage():
    print sys.argv[0], "<dis format database> <proc num>"

def main():
    if len(sys.argv) != 3:
        usage()
        sys.exit(0)
    
    decompile(sys.argv[1], sys.argv[2])

if __name__ == "__main__":
    main()
